#!/bin/bash
##
## Pipeline script for the bwa-mem alignment step. 
## 
##
## Environment variables that should be defined (in job.sh) before calling this script
##  -FastqFolder: Path to folder with FASTQ files
##  -AlignFolder: Folder for output. Will be created or cleared from existing files
##  -REF_FASTA: Path to the reference FASTA file
##  -BWAidx: Path prefix to the index used by bwa-mem
##  -BWA_MEM: Path to bwa-mem program
##  -BwaMemOptions: Other options for the bwa-mem program
##  -ReadGroupStatic: Read group tags that are added to all alignments (eg. CN: and PL: tags)
##  -SAMTOOLS: Path to samtools program
##  -FixmateOptions: Options for the 'samtools fixmate' step
##  -SortOptions: Options for the 'samtools sort' step
##  -CalmdOptions: Options for the 'samtools calmd' step
##  -MergeOptions: Options for the 'samtools merge' step
##  -MarkdupOptions: Options for the 'samtools markdup' step
##  -JAVA: Path to java
##  -GATK: Path to GATK JAR file
##  -DBSNP: Path to VCF file with SNP variants that should be genotyped
##  -HaplotypeCallerOptions: Other options for the the HaplotypeCaller program
##  -TMPDIR: Path to a local temporary working directory
##

set -eo pipefail

## Import utility functions
source ./reggie-utils.sh

## If BWA_MEM is not set, we use BWA_MEM2 by default
if [ -z "${BWA_MEM}" ]; then
	BWA_MEM="${BWA_MEM2}"
fi

## Verify settings in options
rg_var_isdir "FastqFolder" "TMPDIR" "WD"
rg_var_isfile "BWA_MEM" "SAMTOOLS" "JAVA" "PICARD" "GATK" "REF_FASTA" "DBSNP"
rg_var_isset "BWAidx" "AlignFolder" "ReadLength"

## Move to the temporary working directory and create subdirectories
cd ${TMPDIR}
mkdir -p fastq
mkdir -p aligned
mkdir -p merged
mkdir -p stats
mkdir -p tmp/sort
mkdir -p tmp/markdup
mkdir -p tmp/gatk
mkdir -p done

## Read information about paired FASTQ files.
## Columns in 'fastq_info.txt': 
##  -FASTQ_R1
##  -FASTQ_R2
NUM_LINES=`wc -l < ${WD}/fastq_info.txt`

if [ ! -f "done/copy.done" ]; then
	CURRENT_LINE=0
	while IFS=$'\t' read FASTQ1 FASTQ2 READ_GROUP; do
		CURRENT_LINE=$(( ${CURRENT_LINE}+1 ))
		rg_var_isset "FASTQ1" "FASTQ2" "READ_GROUP"
		
		PROGRESS=$(( ${CURRENT_LINE}*8/${NUM_LINES} ))
		rg_progress ${PROGRESS} "Copying FASTQ files: ${CURRENT_LINE} of ${NUM_LINES} pairs"
		
		## Rename FASTQ files so that we don't expose information in the BAM file
		FASTQ1_LOCAL="fastq/${CURRENT_LINE}_R1.fastq.gz"
		FASTQ2_LOCAL="fastq/${CURRENT_LINE}_R2.fastq.gz"
		if [ "${JOB_DEBUG}" ]; then
			# In debug mode we copy only parts of the FASTQ files
			# The ' || true' is critical since 'zcat' will exit with 
			# error when 'head' has closed the pipe
			rg_time "STARTING Copy ${FASTQ1_LOCAL}"
			(zcat ${FastqFolder}/${FASTQ1} || true) | head -n ${MaxFASTQLines} | bgzip -c > ${FASTQ1_LOCAL}
			rg_time "STARTING Copy ${FASTQ2_LOCAL}"
			(zcat ${FastqFolder}/${FASTQ2} || true) | head -n ${MaxFASTQLines} | bgzip -c > ${FASTQ2_LOCAL}
		else
			rg_time "STARTING Copy ${FASTQ1_LOCAL}"
			cp ${FastqFolder}/${FASTQ1} ${FASTQ1_LOCAL}
			rg_time "STARTING Copy ${FASTQ2_LOCAL}"
			cp ${FastqFolder}/${FASTQ2} ${FASTQ2_LOCAL}
		fi
	done < ${WD}/fastq_info.txt

	rg_time "ENDED Copy"
	touch "done/copy.done"
fi

## Align and samtools fixmate, sort and calmd 
if [ ! -f "done/align.done" ]; then
	CURRENT_LINE=0
	while IFS=$'\t' read FASTQ1 FASTQ2 READ_GROUP; do
		CURRENT_LINE=$(( ${CURRENT_LINE}+1 ))

		PROGRESS=$(( 10+${CURRENT_LINE}*30/${NUM_LINES} ))
		rg_progress ${PROGRESS} "Running ${BWA_MEM##*/}: ${CURRENT_LINE} of ${NUM_LINES} FASTQ pairs (${NumThreads} threads)"

		FASTQ1_LOCAL="fastq/${CURRENT_LINE}_R1.fastq.gz"
		FASTQ2_LOCAL="fastq/${CURRENT_LINE}_R2.fastq.gz"
		BAM_OUT="aligned/${CURRENT_LINE}.bam"
	
		rg_time "STARTING ${BWA_MEM##*/} ${BAM_OUT}"
		${BWA_MEM} mem ${BwaMemOptions} -R "@RG\\tID:${CURRENT_LINE}\\t${READ_GROUP}\\t${ReadGroupStatic}" \
			-t ${NumThreads} ${BWAidx} ${FASTQ1_LOCAL} ${FASTQ2_LOCAL} \
			| ${SAMTOOLS} fixmate ${FixmateOptions} --output-fmt bam,level=0 - - \
			| ${SAMTOOLS} sort ${SortOptions} --output-fmt bam,level=0 -T tmp/sort - \
			| ${SAMTOOLS} calmd ${CalmdOptions} --output-fmt bam,level=1 - ${REF_FASTA} \
			> ${BAM_OUT}
		rg_time "ENDED ${BWA_MEM##*/} ${BAM_OUT}"
	
	done < ${WD}/fastq_info.txt
	touch "done/align.done"
fi

## Merge and mark duplicates
if [ ! -f "done/merge.done" ]; then
	rg_progress 50 "Merging ${NUM_LINES} BAM files and checking duplicates"

	rg_time "STARTING merge and index"
	ALL_BAM=`ls -w 0 aligned/*.bam`
	${SAMTOOLS} merge ${MergeOptions} --output-fmt bam,level=0 - ${ALL_BAM} \
		| ${SAMTOOLS} markdup ${MarkdupOptions} \
		-f stats/samtools_markdup.txt \
		--output-fmt bam -T tmp/markdup - - \
		> merged/alignment.bam
	rg_time "ENDED merge"

	rg_progress 60 "Creating BAM index"
	${SAMTOOLS} index -b -@ 8 merged/alignment.bam merged/alignment.bai
	rg_time "ENDED index"

	touch "done/merge.done"
fi

if [ ! -f "done/statistics.done" ]; then
	rg_progress 65 "Calculating alignment statistics"
	rg_time "STARTING statistics"
	
	## Start all steps in the background...
	## Typical time ~15min
	${WD}/alignment_statistics.sh merged/alignment.bam > stats/alignment_statistics.txt &
	ALIGNMENT_STATISTICS_PID=$!

	## Picard: CollectInsertSizeMetrics -- typical time ~25min
	${WD}/stdwrap.sh ${JAVA} ${JavaOptions} \
		-jar ${PICARD} CollectInsertSizeMetrics \
		-INPUT merged/alignment.bam \
		-OUTPUT stats/picard_CollectInsertSizeMetrics.txt \
		-Histogram_FILE stats/picard_CollectInsertSizeMetrics.pdf \
		-QUIET true -VERBOSITY WARNING \
		> picard_CollectInsertSizeMetrics.out &
	INSERT_SIZE_PID=$!

	## Picard: CollectQualityYieldMetrics -- typical time ~30min
	${WD}/stdwrap.sh ${JAVA} ${JavaOptions} \
		-jar ${PICARD} CollectQualityYieldMetrics \
		-INPUT merged/alignment.bam \
		-OUTPUT stats/picard_CollectQualityYieldMetrics.txt \
		-QUIET true -VERBOSITY WARNING \
		> picard_CollectQualityYieldMetrics.out &
	QUALITY_YIELD_PID=$!

	## Picard: CollectAlignmentSummaryMetrics -- typical time ~1h
	${WD}/stdwrap.sh ${JAVA} ${JavaOptions} \
		-jar ${PICARD} CollectAlignmentSummaryMetrics \
		-INPUT merged/alignment.bam \
		-OUTPUT stats/picard_CollectAlignmentSummaryMetrics.txt \
		-QUIET true -VERBOSITY WARNING \
		> picard_CollectAlignmentSummaryMetrics.out &
	ALIGNMENT_SUMMARY_PID=$!

	## Picard: CollectWgsMetrics -- typical time ~1h15min
	${WD}/stdwrap.sh ${JAVA} ${JavaOptions} \
		-jar ${PICARD} CollectWgsMetrics \
		-INPUT merged/alignment.bam \
		-OUTPUT stats/picard_CollectWgsMetrics.txt \
		-R ${REF_FASTA} \
		-READ_LENGTH ${ReadLength} \
		-USE_FAST_ALGORITHM true \
		-QUIET true -VERBOSITY WARNING \
		> picard_CollectWgsMetrics.out &
	WGS_PID=$!
	
	## Wait for the steps to finish in the order they are expected to be completed
	wait ${ALIGNMENT_STATISTICS_PID}
	rg_time "ENDED alignment_statistics.sh"

	rg_progress 70 "Calculating statistics: picard CollectInsertSizeMetrics"
	wait ${INSERT_SIZE_PID}
	rg_time "ENDED picard CollectInsertSizeMetrics"

	rg_progress 75 "Calculating statistics: picard CollectQualityYieldMetrics"
	wait ${QUALITY_YIELD_PID}
	rg_time "ENDED picard CollectQualityYieldMetrics"

	rg_progress 80 "Calculating statistics: picard CollectAlignmentSummaryMetrics"
	wait ${ALIGNMENT_SUMMARY_PID}
	rg_time "ENDED picard CollectAlignmentSummaryMetrics"
	
	rg_progress 85 "Calculating statistics: picard CollectWgsMetrics"
	wait ${WGS_PID}
	rg_time "ENDED picard CollectWgsMetrics"
	rg_time "ENDED statistics"
	
	touch "done/statistics.done"
fi

## Genotype QC
if [ ! -f "done/genotypeqc.done" ]; then
	rg_progress 90 "Running HaplotypeCaller"
	rg_time "STARTING Genotype QC"
	${WD}/stdwrap.sh ${JAVA} ${JavaOptions} \
		-jar ${GATK} HaplotypeCaller \
		-R ${REF_FASTA} \
		-L ${DBSNP} --dbsnp ${DBSNP} --alleles ${DBSNP} \
		--tmp-dir tmp/gatk \
		${HaplotypeCallerOptions} \
		-I merged/alignment.bam \
		-O merged/qc_genotype.vcf \
		> gatk_HaplotypeCaller.out
	rg_time "ENDED Genotype QC"
		
	# Remove '##contig' entries from the qc_genotype.vcf since they are not really needed
	sed -i '/##contig/d' merged/qc_genotype.vcf
	touch "done/genotypeqc.done"
fi


rg_progress 95 "Copying result files to project archive"
rg_time "STARTING Copy"

cp stats/*.txt ${WD}
mkdir -p ${AlignFolder}
rm -rf ${AlignFolder}/*
cp merged/* ${AlignFolder}
cp stats/* ${AlignFolder}
# cp *.out ${AlignFolder}
ls -1 ${AlignFolder}/* > ${WD}/files.out

rg_time "ENDED Copy"
rg_progress 99 "Analysis completed, cleaning up..."
